@@ -73,7 +73,13 @@ def plot_data(hist, ignore_variances=False, color="k", label="Data", ax=None):
 
 
 def plot_signal(
-    hist, label, ignore_variances=False, syst_errs=None, color=None, ax=None
+    hist,
+    label,
+    ignore_variances=False,
+    syst_errs=None,
+    color=None,
+    attach_bands=False,
+    ax=None,
 ):
     """
     Plot signal histogram from PlottableHistogram
@@ -90,6 +96,8 @@ def plot_signal(
         Systematic errors
     color : color
         Line color
+    attach_bands : boolean, optional
+        Attach bands to line in legend. Defaults to False.
     ax : mpl.axes.Axes, optional
         Axes to draw on (defaults to current axes)
     """
@@ -125,7 +133,9 @@ def plot_signal(
                 "Binning mismatch between syst_errs and hist"
             )
         syst_errs = syst_errs_obj.values()
-    _amplplt.plot_signal(label, bins, hist, stat_errs, syst_errs, color, ax)
+    _amplplt.plot_signal(
+        label, bins, hist, stat_errs, syst_errs, color, attach_bands, ax
+    )
 
 
 def plot_ratio(data, total_bkg, ratio_ax, max_ratio=None, plottype="diff"):
@@ -159,25 +169,42 @@ def plot_ratio(data, total_bkg, ratio_ax, max_ratio=None, plottype="diff"):
     if len(data.axes) != 1:
         raise _amplplt.DimensionError("Only 1D histograms are supported here")
     data_obj = data
-    bins = data_obj.axes[0].edges()
+    bins = _bins(data_obj.axes[0])
     data = data_obj.values()
     if data_obj.variances() is None:
         data_errs = _np.sqrt(data)
     else:
         data_errs = _np.sqrt(data_obj.variances())
-    if not isinstance(total_bkg, tuple) or len(total_bkg) != 2:
+
+    if (
+        hasattr(total_bkg, "axes")
+        and hasattr(total_bkg, "values")
+        and hasattr(total_bkg, "variances")
+    ):
+        # total_bkg is a UHI histogram
+        bkg = total_bkg.values()
+        bkg_errs = _np.sqrt(total_bkg.variances())
+    elif isinstance(total_bkg, tuple) and len(total_bkg) == 2:
+        bkg = total_bkg[0]
+        bkg_errs = total_bkg[1]
+    else:
         raise TypeError(
             "total_bkg should be a two element tuple of arrays, as returned by plot_backgrounds"
         )
-    bkg = total_bkg[0]
-    bkg_errs = total_bkg[1]
     _amplplt.plot_ratio(
         bins, data, data_errs, bkg, bkg_errs, ratio_ax, max_ratio, plottype
     )
 
 
 def plot_1d(
-    hist, label, ignore_variances=False, stat_err=True, color=None, ax=None, **kwargs
+    hist,
+    label,
+    ignore_variances=False,
+    stat_err=True,
+    color=None,
+    attach_bands=False,
+    ax=None,
+    **kwargs
 ):
     """
     Plot single 1D histogram from PlottableHistogram
@@ -194,6 +221,8 @@ def plot_1d(
         Draw statistical errors. Defaults to True.
     color : color, optional
         Line color
+    attach_bands : boolean, optional
+        Attach bands to line in legend. Defaults to False.
     ax : mpl.axes.Axes, optional
         Axes to draw on (defaults to current axes)
     **kwargs
@@ -221,7 +250,7 @@ def plot_1d(
             )
     else:
         stat_errs = None
-    _amplplt.plot_1d(label, bins, hist, stat_errs, color, ax, **kwargs)
+    _amplplt.plot_1d(label, bins, hist, stat_errs, color, attach_bands, ax, **kwargs)
 
 
 def plot_2d(hist, ax=None, pad=0.05, **kwargs):
